local scene = {
    loadedScenes = {},
    currentScene = false,
    currentSceneId = false,

    transition = {},
    transitionSwitched = false,
    transitionScene = false,
    transitionSceneData = false,
    currentTransition = false
}
local _scene = {__index = scene}


local fs = love.filesystem
local f = string.format

-- Loads a transition file
function scene:newTransition(id, file)
    assert(self.transition[id] == nil, f("A transition with the id '%s' already exists!", id))
    self.transition[id] = fs.load(file)()
end

-- Loads a scene, But does not switch to it!
function scene:newScene(id, file)
    assert(self.loadedScenes[id] == nil, f("A scene with the id '%s' already exists!", id))
    self.loadedScenes[id] = fs.load(file)()
end

-- Switches to a scene
-- id: The id of the scene as defined in scene:newScene()
-- transition: The id of the transition as defined in scene:newTransition(), If nil or false, No transition will be used
-- data: Any data you want to pass to the new scene
function scene:switch(id, transition, data)
    assert(self.loadedScenes[id], f("Scene '%s' does not exist!", id))
    if transition then
        assert(self.transition[transition], f("Transition '%s' does not exist!", transition))
    end

    if not transition then
        self.currentScene = self.loadedScenes[id]
        self.currentSceneId = id
        if type(self.currentScene.load) == "function" then
            self.currentScene:load(data)
        end
    else
        self.transitionScene = id
        self.transitionSceneData = data
        self.currentTransition = self.transition[transition]
        if type(self.currentTransition.load) == "function" then
            self.currentTransition:load()
        end
    end

end

function scene:inTransition()
    return self.transitionScene and true or false
end

function scene:get()
    return self.currentScene
end

-- Callback functions
function scene:update(dt)
    if type(self.currentScene.update) == "function" then
        self.currentScene:update(dt)
    end

    if self.transitionScene then
        self.currentTransition:update(dt)
        if self.currentTransition.stage == "out" then
            if not self.transitionSwitched then
                self:switch(self.transitionScene, false, self.transitionSceneData)
                self.transitionSwitched = true
            end
        elseif self.currentTransition.stage == "finished" then
            self.transitionScene = false
            self.transitionSceneData = false
            self.transitionSwitched = false
        end
    end
end

function scene:draw()
    if type(self.currentScene.draw) == "function" then
        self.currentScene:draw()
    end

    if self.transitionScene then
        self.currentTransition:draw()
    end
end

function scene:textinput(t)
    if type(self.currentScene.textinput) == "function" then
        self.currentScene:textinput(t)
    end
end

function scene:keypressed(key)
    if type(self.currentScene.keypressed) == "function" then
        self.currentScene:keypressed(key)
    end
end

function scene:keyreleased(key)
    if type(self.currentScene.keyreleased) == "function" then
        self.currentScene:keyreleased(key)
    end
end

function scene:mousepressed(x, y, k)
    if type(self.currentScene.mousepressed) == "function" then
        self.currentScene:mousepressed(x, y, k)
    end

end

function scene:mousereleased(x, y, k)
    if type(self.currentScene.mousereleased) == "function" then
        self.currentScene:mousereleased(x, y, k)
    end
end

function scene:wheelmoved(x, y)
    if type(self.currentScene.wheelmoved) == "function" then
        self.currentScene:wheelmoved(x, y)
    end

end

function scene:resize(width, height)
    if type(self.currentScene.resize) == "function" then
        self.currentScene:resize(width, height)
    end
end

return scene